/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"
#include "toolbar.h"
#include "sort.h"
#include "rfc822.h"

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <errno.h>

#define msgno(X) *Context->v2r[X]

#define REDRAW_FULL 1
#define REDRAW_INDEX 2
#define REDRAW_CURRENT 4
#define REDRAW_MOTION 8
#define REDRAW_STATUS 16

#define HDR_OFFSET 1
#define HDR_CMD (LINES-1)
#define HDR_LEN (HDR_CMD-HDR_OFFSET-1)

int ci_sigwinch = 0;
jmp_buf Jumpbuf;

/* The folder the user last saved to.  Used by ci_save_message() */
static char ci_last_folder[_POSIX_PATH_MAX] = { 0 };

#ifdef NCURSES_MOUSE_VERSION
TOOLBAR Main_toolbar[] = {
  { "Quit",  OP_MAIN_QUIT, 0 },
  { "Top",   OP_MAIN_FIRST_MESSAGE, 0 },
  { "Bot",   OP_MAIN_LAST_MESSAGE, 0 },
  { "NxtPg", OP_MAIN_NEXT_PAGE, 0 },
  { "PrvPg", OP_MAIN_PREVIOUS_PAGE, 0 },
  { "Del",   OP_MAIN_DELETE_MESSAGE, 0 },
  { "UnDel", OP_MAIN_UNDELETE_MESSAGE, 0 },
  { "Attch", OP_MAIN_VIEW_ATTACHMENTS, 0 },
  { "Mail",  OP_MAIN_MAIL, 0 },
  { "Reply", OP_MAIN_REPLY, 0 },
  { "Group", OP_MAIN_GROUP_REPLY, 0 },
  { "Forw",  OP_MAIN_FORWARD_MESSAGE, 0 },
  { 0, 0, 0 }
};
#endif

/*
 * Better display with 8-bit characters - pms@romance.kaist.ac.kr
 *
 * force_subj is used to force printing of the subject in threading mode
 * so that the subject is displayed when threads overlap a page boundary
 */
static void ci_generate_header (char *s, int l, HEADER *h, int force_subj)
{
  int len;
  int cols = option(OPTARROWCURSOR) ? COLS-2 : COLS;
  register int i, ch8bit;

  mutt_generate_header(s, l, h, force_subj);
  len = strlen(s);
  if (len < cols-1)
    sprintf(s+len, "%*.*s", cols-len-1, cols-len-1, " ");
  else {
     for(i=ch8bit=0; i<cols-1; i++) {
         if(s[i] & 0x80) ch8bit++; 
         else ch8bit = 0;
     }
     if((i-ch8bit)%2==0) s[cols-2] = 0x20;          
  }
  s[cols-1] = 0;
}

static void ci_show_last_error (void)
{
  CLEARLINE(LINES-1);
  SETCOLOR(MT_COLOR_ERROR);
  addstr(Errorbuf);
  SETCOLOR(MT_COLOR_NORMAL);
}

#ifdef METAMAIL
static int ci_invoke_metamail (HEADER *h, FILE *f)
{
  long chunk, bytes = h->length;
  FILE *fout;
  char tempfile[_POSIX_PATH_MAX], buffer[LONG_STRING];

  mutt_mktemp(tempfile);
  if ((fout = fopen(tempfile, "w")) == NULL) return -1;
  if (fseek(f, h->offset, 0) == -1) {
    fclose(fout);
    return -1;
  }
  while (bytes > 0) {
    chunk = (bytes > sizeof(buffer)) ? sizeof(buffer) : bytes;
    chunk = fread(buffer, 1, chunk, f);
    fwrite(buffer, 1, chunk, fout);
    bytes -= chunk;
  }
  fclose(fout);
  snprintf(buffer, sizeof(buffer), "%s -P -z %s", METAMAIL, tempfile);
  endwin();
  mutt_system(buffer);
  return 0;
}
#endif /* METAMAIL */

static int ci_display_message (HEADER *cur)
{
  char tempfile[_POSIX_PATH_MAX], buffer[STRING];
  char *cp;
  STATE *st;
  int rc=0, builtin=0;
  int len;

#ifdef METAMAIL
  if (cur->metamail) {
    ci_invoke_metamail(cur, Context->folder->fp);
    goto mark_it_read;
  }
#endif

  mutt_parse_mime_message(cur);

#ifdef _PGPPATH
  /* see if PGP is needed for this message.  if so, we should exit curses */
  if (cur->pgp) {
    if (cur->pgp == PGPENCRYPT && !pgp_valid_passphrase()) return 0;
    endwin();
  }
#endif

  st = mutt_new_state();
  st->displaying = 1;

  if (strcmp(Pager, "builtin") == 0) {
    mutt_mktemp(tempfile);
    if ((st->fpout = fopen(tempfile, "w")) == NULL) {
      mutt_error("Could not create temporary file!");
      return(0);
    }
    builtin++;
  }
  else {
    if (!isendwin()) endwin();
    mutt_create_filter(Pager, &(st->fpout), 0);
  }

  if (cur->deleted)
    cp = "[Deleted]";
  else if (cur->read) {
    if (cur->tagged)
      cp = "[Tagged]";
    else if (cur->flagged)
      cp = "[Flagged]";
    else
      cp = "Message";
  }
  else if (cur->old)
    cp = "[Old]";
  else
    cp = "[New]";

  snprintf(buffer, sizeof(buffer), "%s %d/%d: ", cp, cur->msgno,
	   Context->msgcount);
  fputs(buffer, st->fpout);
  len = strlen(buffer);

  if (cur->from && cur->from->personal)
    fprintf(st->fpout, "%-*.*s\n", COLS-len, COLS-len, cur->from->personal);
  else {
    buffer[0] = 0;
    mutt_write_address(buffer, sizeof(buffer) - strlen(buffer), cur->from);
    fprintf(st->fpout, "%-*.*s\n", COLS-len, COLS-len, buffer);
  }

  if (cur->subject) {
    int len = strlen(cur->subject);

    len = (COLS-len)/2;
    if (len < 0) {
      /* The subject is longer than the current line! */
      fprintf(st->fpout, "%*.*s", COLS, COLS, cur->subject);
    }
    else
      fprintf(st->fpout, "%*.*s%s", len, len, "", cur->subject);
  }
  fputs("\n\n", st->fpout);

  st->fpin = Context->folder->fp;
  mutt_copy_header(Context->folder->fp, cur, st->fpout, (Context->weed ? CH_WEED : 0)|CH_DECODE);
  mutt_body_handler(cur->content, st);
  fclose(st->fpout);
  mutt_free_state(&st);

  if (builtin) {
    /* Invoke the builtin pager */
    rc = ci_simple_pager(tempfile, TRUE); /* removes file upon completion */
  }
  else {
    wait (&rc);
    keypad(stdscr, TRUE);
    if (option(OPTPROMPTAFTER)) {
      rc = ci_any_key_to_continue("Command: ");
      rc = keymap_lookup(rc, MENU_PAGER);
    }
    else
      rc = 0;
  }

#ifdef METAMAIL
mark_it_read:
#endif
  /* mark the message as read and update the new count, if needed */
  if (!cur->read) {
    cur->read = 1;
    cur->changed = 1;
    if (!cur->old) Context->new--;
  }

  return(rc);
}

static int ci_next_undeleted (int msgno)
{
  int i;

  for (i=msgno+1; i<Context->vcount; i++)
    if (! Context->hdrs[*Context->v2r[i]]->deleted)
      return(i);
  return(-1);
}

static int ci_previous_undeleted (int msgno)
{
  int i;

  for (i=msgno-1; i>=0; i--)
    if (! Context->hdrs[*Context->v2r[i]]->deleted)
      return(i);
  return(-1);
}

/* returns 1 if OK to proceed, 0 to abort */
static int confirm_func (const char *s)
{
  char msg[STRING];
  struct stat st;
  int ret=1;
  int is_mail_dir = strncmp(s, Maildir, strlen(Maildir)) == 0 ? 1 : 0;
  int is_folder=0;
  FILE *f;

  if (stat(s, &st) != -1) {
    if (is_mail_dir) {
      if ((f = fopen(s, "r")) == NULL) {
	mutt_error("Permission denied.");
	ret=0;
	goto finish;
      }
      fgets(msg, sizeof(msg), f);
      if (strncmp("From ", msg, 5) == 0) is_folder=1;
    }
    if (option(OPTCONFIRMAPPEND) || (!is_folder && option(OPTCONFIRMFILES))) {
      snprintf(msg, sizeof(msg), "Append messages to %s %s?",
	       is_folder ? "folder" : "file", s);
      if (mutt_yesorno(msg, 1) == 0) ret=0;
    }
  }
  else if (errno == ENOENT) {
    if (option(OPTCONFIRMCREATE) || (is_mail_dir && option(OPTCONFIRMFOLDERS))) {
      snprintf(msg, sizeof(msg), "Create %s %s?", is_mail_dir ? "folder" : "file", s);
      if (mutt_yesorno(msg, 1) == 0) ret=0;
    }
  }
  else {
    mutt_error("Could not stat %s!", s);
    ret=0;
  }

finish:
  CLEARLINE(LINES-1);
  return(ret);
}

/* returns 0 if the copy/save was successful, or -1 on error/abort */
static int ci_save_message (HEADER *h, int delete, int ignore_selection,
			    int *redraw)
{
  int ch;
  char buffer[_POSIX_PATH_MAX];

  *redraw = 0;
  mutt_default_savename(buffer, sizeof(buffer), h);
  CLEARLINE(LINES-1);
  printw("%s to folder (? for list): %s", delete ? "Save" : "Copy", buffer);
  refresh();
  ch = ci_getch();
  if (ch == '?') {    
    *redraw = REDRAW_FULL;
    folder_select(Maildir, buffer);
    if (buffer[0] == 0) return(-1);
  }
  else if (ch == ctrl('G')) {
    CLEARLINE(LINES-1);
    return(-1);
  }
  else if (!CI_is_return(ch)) {
    if (IsPrint(ch)) buffer[0] = 0;
    ungetch(ch);
    if (ci_get_field(delete ? "Save to folder: " : "Copy to folder: ", buffer, sizeof(buffer), CI_complete) == -1 || !buffer[0])
      return(-1);
  }
  
  /*
   * This is an undocumented feature of ELM pointed out to me by Felix von
   * Leitner <leitner@prz.fu-berlin.de>
   */
  if (strcmp(buffer, ".") == 0)
    strfcpy(buffer, ci_last_folder, sizeof(buffer));
  else
    strfcpy(ci_last_folder, buffer, sizeof(ci_last_folder));

  mutt_expand_path(buffer);

  /* check to make sure that this file is really the one the user wants */
  if (confirm_func(buffer) == 0) {
    CLEARLINE(LINES-1);
    return(-1);
  }

  if (!ignore_selection && Context->tagged)
    mutt_save_message(0, buffer, delete);
  else
    mutt_save_message(h, buffer, delete);

  return(0);
}

static void ci_bounce_message (HEADER *h)
{
  char prompt[SHORT_STRING];
  char buf[LONG_STRING] = { 0 };
  ADDRESS *adr = 0, *padr;

  FOREVER {
    if (ci_get_field("Bounce message to: ", buf, sizeof(buf), 0) != 0 || !buf[0])
      return;
    if (mutt_parse_adrlist(&adr, buf, "@") == 0)
      break;
    mutt_free_address(&adr);
  }
  adr = mutt_expand_aliases(adr);

  buf[0] = 0;
  mutt_write_address(buf, sizeof(buf), adr);
  buf[(COLS+1)/2]=0; /* so we have a nicer display */
  snprintf(prompt, sizeof(prompt), "Bounce message to %s...?", buf);
  if (!mutt_yesorno(buf, 1)) {
    CLEARLINE(LINES-1);
    return;
  }

  buf[0] = 0;
  padr = adr;
  while (padr) {
    strcat(buf, " ");
    strcat(buf, padr->mailbox);
    if (*padr->host != '@') {
      strcat(buf, "@");
      strcat(buf, padr->host);
    }
    padr = padr->next;
  }
  mutt_free_address(&adr);

  if (Context->tagged)
    mutt_bounce_message(NULL, buf);
  else
    mutt_bounce_message(h, buf);
  mutt_error("Message%s bounced.", Context->tagged ? "s" : "");
}

static int ci_pipe_message (HEADER *h)
{
  char buffer[LONG_STRING];

  buffer[0] = 0;
  if (ci_get_field("Pipe to command: ", buffer, sizeof(buffer), 0) != 0 || !buffer[0])
    return 0;
  endwin();
  mutt_pipe_message(h, buffer);
  ci_any_key_to_continue(0);
  return 1;
}

static int ci_select_sort_method (void)
{
  int method = Sort; /* save the current method in case of abort */
  int ch;

  Sort = 0;
  while (!Sort) {
    mvaddstr(LINES-1,0,"Sort by (d)ate/(f)rom/(o)rder/(r)eceived/(s)ubj/(t)hreads/si(z)e?: ");
    ch = ci_getch();
    if (ch == ctrl('G')) {
      Sort = method;
      CLEARLINE(LINES-1);
      return(-1);
    }
    switch (ch) {
    case 'D':
      Sort = SORTREVERSE;
    case 'd':
      Sort |= SORTDATE;
      break;
    case 'F':
      Sort = SORTREVERSE;
    case 'f':
      Sort |= SORTFROM;
      break;
    case 'O':
      Sort = SORTREVERSE;
    case 'o':
      Sort |= SORTORDER;
      break;
    case 'S':
      Sort = SORTREVERSE;
    case 's':
      Sort |= SORTSUBJECT;
      break;
    case 't':
    case 'T':
      Sort = SORTTHREADS;
      break;
    case 'R':
      Sort = SORTREVERSE;
    case 'r':
      Sort = SORTRECEIVED;
      break;
    case 'Z':
      Sort = SORTREVERSE;
    case 'z':
      Sort |= SORTSIZE;
      break;
    }
  }
  CLEARLINE(LINES-1);
  return(Sort != method ? 0 : -1); /* no need to resort if it's the same */
}

/*
 * Return the index of the first new message, or failing that, the first
 * unread message.
 */
static int ci_first_message (void)
{
  int old = -1, i;

  for (i=0; i < Context->vcount; i++) {
    if (! Context->hdrs[*Context->v2r[i]]->read && ! Context->hdrs[*Context->v2r[i]]->deleted) {
      if (! Context->hdrs[*Context->v2r[i]]->old)
	return(i);
      else if (old == -1)
	old = i;
    }
  }
  return(old);
}

#define print_enriched_string(X) real_print_enriched_string(X,1)

static void real_print_enriched_string (unsigned char *s, int do_color)
{
  if (Sort != SORTTHREADS) {
    /* Don't need to do any work if we aren't using threads! */
    addstr(s);
    return;
  }
  
  while (*s) {
    if (*s < ctrl('G')) {
      if (do_color) SETCOLOR(MT_COLOR_TREE);
      while (*s && *s < ctrl('G')) {
	switch(*s) {
	case 1:
	  addch(ACS_LLCORNER);
	  break;
	case 2:
	  addch(ACS_LTEE);
	  break;
	case 3:
	  addch(ACS_HLINE);
	  break;
	case 4:
	  addch(ACS_VLINE);
	  break;
	case 5:
	  addch(' ');
	  break;
	case 6:
	  addch('>');
	  break;
	}
	s++;
      }
      if (do_color) SETCOLOR(MT_COLOR_NORMAL);
    }
    else {
      addch(*s);
      s++;
    }
  }
}

/*
 * This function handles the message index window as well as commands returned
 * from the pager (MENU_PAGER).
 */
void ci_main (void)
{
  volatile int top = 0;
  volatile int oldcurrent = Context->current;
  volatile int op;

  int redraw = REDRAW_FULL;
  int menu = MENU_MAIN; /* which menu are we currently in? */
  int done = 0;
  int i = 0;
  char buffer[LONG_STRING];
  char errbuf[SHORT_STRING];

#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) toolbar_init(Main_toolbar);
#endif

  if (Context->msgcount && option(OPTPOINTNEW)) {
    if ((Context->current = ci_first_message()) == -1)
      Context->current = Sort & SORTREVERSE ? 0 : Context->vcount - 1;
  }
  else
    Context->current = 0;
  Context->weed = 1; /* header weeding is on to begin with */

/* Consider newbie for Mutt - pms@romance.kaist.ac.kr */
  mutt_error("   ?-help /-search, d)elete, f)orward, m)ail, r)eply, s)ave, u)ndelete, e(x)it"); 

  FOREVER {
    if (ci_sigwinch) {
      endwin();
      initscr();
      refresh();
      
      ci_sigwinch = 0;
      redraw = REDRAW_FULL;
      menu = MENU_MAIN;
    }
    
    if (menu == MENU_MAIN) {
      /* check for new mail in the mailbox */
      if (mutt_check_mailbox(Context)) {
	mutt_sort_headers(Context);
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      
      if (redraw == REDRAW_FULL) {
	clear();
#ifdef NCURSES_MOUSE_VERSION
	if (Mousemask) toolbar_post(Main_toolbar);
#endif
	ci_show_last_error();
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      }

      if (redraw & REDRAW_STATUS) {
	mutt_status_line(buffer, sizeof(buffer));
	if (option(OPTXTERMMOUSE) || !option(OPTSTATUSONTOP))
	  CLEARLINE(LINES-2);
	else
	  CLEARLINE(0);
	SETCOLOR(MT_COLOR_STATUS);
	printw("%-*.*s", COLS, COLS, buffer);
	SETCOLOR(MT_COLOR_NORMAL);
	redraw &= ~REDRAW_STATUS;
      }

      if (Context && Context->folder) {
	if (Context->current >= top + HDR_LEN) {
	  while (top + HDR_LEN  - 1 < Context->current) top += HDR_LEN;
	  redraw = REDRAW_INDEX;
	}
	else if (Context->current < top) {
	  while (top > Context->current) top -= HDR_LEN;
	  if (top < 0) top = 0;
	  redraw = REDRAW_INDEX;
	}
	else if (oldcurrent != Context->current) {
	  if (!redraw) redraw = REDRAW_MOTION;
	}
      }

      if (redraw == REDRAW_INDEX) {
	for (i=top; i<top+HDR_LEN; i++) {
	  if (i < Context->vcount) {
	    ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[i]], (i==top));
	    CLEARLINE(i-top+HDR_OFFSET);
	    
	    if (option(OPTARROWCURSOR))
	      move(i-top+HDR_OFFSET, 2);
	    else if (i == Context->current)
	      SETCOLOR(MT_COLOR_INDICATOR);
	    real_print_enriched_string(buffer, i!=Context->current || option(OPTARROWCURSOR));
	    if (!option(OPTARROWCURSOR) && i==Context->current)
	      SETCOLOR(MT_COLOR_NORMAL);
	  }
	  else
	    CLEARLINE(i-top+HDR_OFFSET);
	}
	redraw = 0;
      }
      else if (redraw == REDRAW_MOTION) {
	if (option(OPTARROWCURSOR)) {
	  ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[oldcurrent]], (oldcurrent==top));
	  mvaddstr(oldcurrent-top+HDR_OFFSET, 0, "  ");
	  print_enriched_string(buffer);
	}
	else {
	  /* redraw the current line in case the flags changed. */
	  CLEARLINE(oldcurrent-top+HDR_OFFSET);
	  ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[oldcurrent]], (oldcurrent == top));
	  print_enriched_string(buffer);
		  
	  move(Context->current-top+HDR_OFFSET, 0);
      ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[Context->current]], (Context->current == top));
	  SETCOLOR(MT_COLOR_INDICATOR);
	  real_print_enriched_string(buffer, 0);
	  SETCOLOR(MT_COLOR_NORMAL);
	}
	redraw = 0;
      }
      else if (redraw == REDRAW_CURRENT) {
	if (option(OPTARROWCURSOR))
	  move(Context->current-top+HDR_OFFSET, 2);
	else {
	  move(Context->current-top+HDR_OFFSET, 0);
	  SETCOLOR(MT_COLOR_INDICATOR);
	}	      
	ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[Context->current]], (Context->current == top));
	real_print_enriched_string(buffer, option(OPTARROWCURSOR));
	if (!option(OPTARROWCURSOR)) SETCOLOR(MT_COLOR_NORMAL);
	redraw = 0;
      }
      
      if (Context) {
	oldcurrent = Context->current;
	if (option(OPTARROWCURSOR)) {
	  move(Context->current-top+HDR_OFFSET, 0);
	  if (Context->msgcount) {
	    SETCOLOR(MT_COLOR_INDICATOR);
	    addstr("->");
	    SETCOLOR(MT_COLOR_NORMAL);
	  }
	}
	else
	  move(Context->current-top+HDR_OFFSET, COLS-1);
      }
      else
	move(LINES-1, COLS-1);
      
      refresh();
      
      if (Timeout) {
	timeout(Timeout*1000); /* milliseconds */
      }
      
      if ((i = ci_getch()) == ERR) continue;
	  
      if (Timeout) {
	timeout(-1); /* restore blocking operation */
      }
	  
      ci_clear_error();
	  
#ifdef NCURSES_MOUSE_VERSION
      if (i == KEY_MOUSE) {
	MEVENT e;

	dprint(3, (debugfile, "ci_main(): detected mouse event.\n"));

	getmouse(&e);
	if (e.y == 0) {
	  /* This click occurred on the toolbar. */
	  if (e.bstate & BUTTON1_CLICKED) {
	    if ((op = toolbar_select(Main_toolbar, e.x)) == -1) continue;
	  }
	}
	else if (e.bstate & BUTTON1_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    Context->current = e.y - HDR_OFFSET + top;
	    if (Context->current > Context->vcount - 1)
	      Context->current = oldcurrent;
	  }
	  continue;
	}
	else if (e.bstate & BUTTON1_DOUBLE_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    Context->current = e.y - HDR_OFFSET + top;
	    if (Context->current > Context->vcount - 1) {
	      Context->current = oldcurrent;
	      continue;
	    }
	    op = ci_display_message(Context->hdrs[*Context->v2r[Context->current]]);
	    menu = MENU_PAGER;
	  }
	  continue;
	}
      }
      else
#endif /* NCURSES_MOUSE_VERSION */
	op = keymap_lookup(i, MENU_MAIN);
    } /* if (menu == MENU_MAIN) */
    
    switch (op) {
    case OP_MAIN_BOTTOM_OF_PAGE:
      if (Context->vcount) {
	Context->current = top + HDR_LEN - 1;
	if (Context->current > Context->vcount - 1)
	  Context->current = Context->vcount - 1;
      }
      break;
	  
    case OP_MAIN_DELETE_PATTERN:
      if (Context->vcount) {
	ci_pattern_function(ACTIONDELETE, "Delete messages matching: ");
	redraw = REDRAW_INDEX;
      }
      break;
      
    case OP_MAIN_ENTER_COMMAND:
      buffer[0] = 0;
      if (ci_get_field(":", buffer, sizeof(buffer), 0) != 0 || !buffer[0]) break;
      mutt_parse_rc_line(buffer, 0, errbuf, sizeof(errbuf));
      if (errbuf[0]) {
	/*
	 * since errbuf could potentially contain printf() sequences in it,
	 * we must call mutt_error() in this fashion so that vsprintf()
	 * doesn't expect more arguments that we passed
	 */
	mutt_error("%s", errbuf);
      }
      else {
	/* might have performed some action that changed the screen */
	redraw = REDRAW_FULL;
      }
      break;

#ifdef USE_POP
    case OP_MAIN_FETCH_MAIL:
      mutt_fetchPopMail();
      redraw = REDRAW_FULL;
      break;
#endif /* USE_POP */
      
    case OP_MAIN_FIRST_MESSAGE:
      Context->current = 0;
      if (top != 0) {
	top = 0;
	redraw = REDRAW_INDEX;
      }
      break;
      
    case OP_MAIN_HELP:
      ci_help(MENU_MAIN);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_JUMP:
      if (Context->vcount) {
	buffer[0] = 0;
        ungetch(i);
	if (ci_get_field("Jump to message: ", buffer, sizeof(buffer), 0) != 0 || !buffer[0])
	  break;
	
	if (! isdigit(buffer[0])) {
	  mutt_error("Argument must be a message number.");
	  break;
	}
	
	i = atoi(buffer);
	if (i > 0 && i <= Context->msgcount) {
	  if (Context->hdrs[i-1]->virtual != -1)
	    Context->current = Context->hdrs[i-1]->virtual;
	  else
	    mutt_error("That message is not visible.");
	}
	else
	  mutt_error("Invalid message number.");
      }
      break;
	  
    case OP_MAIN_JUMP_NEW:
      if (Context->vcount) {
	if ((Context->current = ci_first_message()) == -1) {
          Context->current = oldcurrent;
	  if (menu == MENU_MAIN) mutt_error("No new messages.");
	}
	else if (menu == MENU_PAGER) {
	  op = OP_MAIN_DISPLAY_MESSAGE;
	  continue;
	}
      }
      break;

    case OP_MAIN_JUMP_UNREAD:
      if (!Context->vcount) break;
      for (i=0; i<Context->vcount; i++) {
	if (! Context->hdrs[*Context->v2r[i]]->read) {
	  Context->current = i;
	  break;
	}
      }
      if (i == Context->vcount) mutt_error("No unread messages.");
      break;
	  
    case OP_MAIN_LAST_MESSAGE:
      if (Context->vcount) {
	Context->current = Context->vcount - 1;
	while (top + HDR_LEN < Context->current) top += HDR_LEN;
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_LIMIT:
      if (Context->vcount) {
    if (ci_pattern_function(ACTIONLIMIT, "Limit to messages matching: ") == 0)
	  Context->current = 0;
    redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      break;	  
      
    case OP_MAIN_MIDDLE_OF_PAGE:
      if (Context->vcount) {
	i = top + HDR_LEN;
	if (i > Context->vcount - 1)
	  i = Context->vcount - 1;
	Context->current = top + (i - top) / 2;
      }
      break;
      
    case OP_MAIN_NEXT_PAGE:
      if (Context->vcount) {
	if (top + HDR_LEN < Context->vcount) {
	  top += HDR_LEN;
	  Context->current = top;
	  redraw = REDRAW_INDEX;
	}
	else
	  mutt_error("You are on the last page.");
      }
      break;
	  
    case OP_MAIN_PREVIOUS_PAGE:
      if (Context->vcount) {
	if (top > 0) {
	  top -= HDR_LEN;
	  if (top < 0) top  = 0;
	  Context->current = top;
	  redraw = REDRAW_INDEX;
	}
	else
	  mutt_error("You are on the first page.");
      }
      else
	beep();
      break;
      
    case OP_MAIN_QUIT:
      if (mutt_close_mailbox(Context))
	done = 1;
      else
	redraw = REDRAW_FULL; /* new mail arrived? */
      break;
      
    case OP_MAIN_RECALL_MESSAGE:
      ci_send_message(0,0,SENDPOSTPONED);
      redraw = REDRAW_FULL;
      break;

    case OP_MAIN_REFRESH:
      redraw = REDRAW_FULL;
      break;
	  
    case OP_MAIN_SEARCH:
      if (!Context->vcount) break;
      
      if ((Context->current = ci_search_command(Context->current)) == -1)
	Context->current = oldcurrent;
      
      break;

    case OP_MAIN_SORT_MESSAGES:
      if (ci_select_sort_method() == 0) {
	if (Context->vcount) {
	  mutt_sort_headers(Context);
	  redraw = REDRAW_INDEX | REDRAW_STATUS;
	}
      }
      break;
	  
    case OP_MAIN_SUBSHELL:
      CLEARLINE(LINES-1);
      endwin();
      puts("Type \"exit\" to return to Mutt.");
      fflush(stdout);
      mutt_system(Shell);
      break;      
      
    case OP_MAIN_SYNC_FOLDER:
      if (! Context->vcount) break;

      if (! mutt_check_folder(Context)) {
	mutt_error("Mailbox was not modified.");
	break;
      }
      
      strfcpy(buffer, Context->folder->name, sizeof(buffer));
      if (mutt_sync_mailbox(Context)) {
	if ((Context = mutt_open_mailbox(buffer, Context, 0)) == NULL) {
	  /* User must have deleted all the messages, so just exit. */
	  done = 1;
	  break;
	}
    if (Context->msgcount && option(OPTPOINTNEW)) {
	  if ((Context->current = ci_first_message()) == -1)
	    Context->current = 0;
	}
	else
	  Context->current = 0;
	redraw = REDRAW_FULL;
      }
      else
	redraw = REDRAW_FULL; /* new mail arrived? */
      break;
            
    case OP_MAIN_TAG_MESSAGE:
      if (Context->vcount) {
	if (Context->hdrs[msgno(Context->current)]->tagged) {
	  Context->hdrs[msgno(Context->current)]->tagged = 0;
	  Context->tagged--;
	}
	else {
	  Context->hdrs[msgno(Context->current)]->tagged = 1;
	  Context->tagged++;
	}
	redraw = REDRAW_STATUS;
	if (Context->current < Context->vcount - 1)
	  Context->current++;
	else
	  redraw |= REDRAW_CURRENT;
      }
      break;
      
    case OP_MAIN_TAG_PATTERN:
      if (Context->vcount) {
	ci_pattern_function(ACTIONSELECT, "Tag messages matching: ");
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      break;
	  
    case OP_MAIN_TOP_OF_PAGE:
      if (Context->vcount)
	Context->current = top;
      break;
      
    case OP_MAIN_UNDELETE_PATTERN:
      if (Context->vcount) {
	if (ci_pattern_function(ACTIONUNDELETE, "Undelete messages matching: ") == 0)
	  redraw = REDRAW_INDEX;
      }
      break;
      
    case OP_MAIN_UNTAG_PATTERN:
      if (Context->vcount) {
	if (ci_pattern_function(ACTIONUNSELECT, "Untag messages matching: ") == 0)
	  redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      break;
      
    case OP_MAIN_UNDELETE_MESSAGE:
      if (! Context->vcount) break;

      Context->hdrs[msgno(Context->current)]->deleted = 0;
      if (Context->current < Context->vcount-1)
	Context->current++;
      else
	redraw = REDRAW_CURRENT;
      break;

    case OP_MAIN_VIEW_ATTACHMENTS:
      if (! Context->vcount) break;

      if (Context->hdrs[msgno(Context->current)]->content->type == TYPEMULTIPART) {
	mutt_parse_mime_message(Context->hdrs[*Context->v2r[Context->current]]);	
	ci_attach(Context->hdrs[*Context->v2r[Context->current]]->content);
	redraw = REDRAW_FULL;
      }
      else
	mutt_error("This is not a multipart message.");
      break;
            
    case OP_MAIN_NEXT_THREAD:
      if (Context->vcount) {
	if ((Context->current = mutt_next_thread(Context->hdrs[*Context->v2r[Context->current]])) == -1) {
	  Context->current = oldcurrent;
	  mutt_error("No more threads.");
	}
      }
      break;

    case OP_MAIN_PREVIOUS_THREAD:
      if (Context->vcount) {
	if ((Context->current = mutt_previous_thread(Context->hdrs[*Context->v2r[Context->current]])) == -1) {
	  Context->current = oldcurrent;
	  mutt_error("You are on the first thread.");
	}
      }
      break;

    case OP_MAIN_EXECUTE_MACRO:
      mutt_execute_macro(MENU_MAIN, i);
      break;

      /* ----------------------------------------------------------------------
       * The following operations can be performed inside of the pager.
       */
      
    case OP_MAIN_ALIAS:
      if (Context->vcount)
	ci_mkalias(Context->hdrs[*Context->v2r[Context->current]]);
      break;
	  
    case OP_MAIN_BOUNCE_MESSAGE:
      if (Context->vcount)
	ci_bounce_message(Context->hdrs[*Context->v2r[Context->current]]);
      break;
	  
    case OP_MAIN_CHANGE_FOLDER:
      mvaddstr(LINES-1, 0, "Open folder ('?' for list): ");
      refresh();
      i = ci_getch();
      CLEARLINE(LINES-1);
      if (CI_is_return(i) || i == ctrl('G')) break;
      if (i != '?') {
        ungetch(i);
	buffer[0] = 0;
	if (ci_get_field("Open folder: ", buffer, sizeof(buffer), CI_complete) != 0 || !buffer[0])
	  break;
      }
      else {
	refresh();
	folder_select(Maildir, buffer);
	redraw = REDRAW_FULL;
	if (!buffer[0]) break;
      }
      mutt_expand_path(buffer);
      if (!mutt_is_valid_mailbox(buffer)) break;
      if (!mutt_close_mailbox(Context)) break;
      
      if ((Context = mutt_open_mailbox(buffer, Context, 0)) == NULL) {
	mutt_error("Unable to open %s!", buffer);
	sleep(1); /* pause before exiting */
	done = 1;
	break;
      }
      if (Context->msgcount && option(OPTPOINTNEW)) {
	if ((Context->current = ci_first_message()) == -1)
	  Context->current = Sort & SORTREVERSE ? 0 : Context->vcount - 1;
      }
      else
	Context->current = 0;
      ci_clear_error();
      redraw = REDRAW_FULL;
      break;
	  
    case OP_MAIN_DELETE_THREAD:
      if (Context->vcount) {
	mutt_delete_thread(Context->hdrs[*Context->v2r[Context->current]]);
	if ((Context->current = ci_next_undeleted(Context->current)) == -1)
	  Context->current = oldcurrent;
	else if (menu == MENU_PAGER) {
	  op = OP_MAIN_DISPLAY_MESSAGE;
	  continue;
	}
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_DISPLAY_MESSAGE:
    case OP_MAIN_DISPLAY_HEADERS: /* don't weed the headers */
      if (Context->vcount) {
	/*
	 * toggle the weeding of headers so that a user can press the key
	 * again while reading the message.
	 */
	if (op == OP_MAIN_DISPLAY_HEADERS)
	  Context->weed = !Context->weed;
	
	op = ci_display_message(Context->hdrs[*Context->v2r[Context->current]]);
	menu = MENU_PAGER;
	oldcurrent = Context->current;
	continue;
      }
      break;
      
    case OP_MAIN_EXIT:
      if (menu == MENU_MAIN) {
	mutt_fastclose_mailbox(Context);
	done = 1;
      }
      break;
	  
    case OP_MAIN_NEXT_UNDELETED:
      if (! Context->vcount) break;
      if (Context->current >= Context->vcount - 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the last message.");
	break;
      }
      if ((Context->current = ci_next_undeleted(Context->current)) == -1) {
	Context->current = oldcurrent;
	mutt_error("No undeleted messages.");
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
      
    case OP_MAIN_NEXT_MESSAGE:
      if (! Context->vcount) break;
      if (Context->current >= Context->vcount - 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the last message.");
	break;
      }
      Context->current++;
      if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_PREVIOUS_UNDELETED:
      if (! Context->vcount) break;
      if (Context->current < 1) {
	mutt_error("You are on the first message.");
	break;
      }
      if ((Context->current = ci_previous_undeleted(Context->current)) == -1) {
	Context->current = oldcurrent;
	mutt_error("No undeleted messages.");
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_PREVIOUS_MESSAGE:
      if (! Context->vcount) break;
      if (Context->current < 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the first message.");
	break;
      }
      Context->current--;
      if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_COPY_MESSAGE:
    case OP_MAIN_SAVE_MESSAGE:
      if (Context->vcount) {
	if (ci_save_message(Context->hdrs[*Context->v2r[Context->current]],
			    (op == OP_MAIN_SAVE_MESSAGE),
			    0,
			    &redraw) != 0 || op != OP_MAIN_SAVE_MESSAGE)
	  break;

	if (!Context->tagged) {
	  if ((Context->current = ci_next_undeleted(Context->current)) == -1) {
	    Context->current = oldcurrent;
	    if (!redraw) redraw = REDRAW_CURRENT;
	  }
	  else if (menu == MENU_PAGER) {
	    op = OP_MAIN_DISPLAY_MESSAGE;
	    continue;
	  }
	  else if (!redraw)
	    redraw = REDRAW_INDEX;
	}
	else if (!redraw)
	  redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_DELETE_MESSAGE:
      if (!Context->vcount) break;

      Context->hdrs[msgno(Context->current)]->deleted = 1;
      if ((Context->current = ci_next_undeleted(Context->current)) == -1) {
	Context->current = oldcurrent;
	if (menu == MENU_MAIN) redraw = REDRAW_CURRENT;
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
      
    case OP_MAIN_FORWARD_MESSAGE:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDFORWARD);
	redraw = REDRAW_FULL;
      }
      break;
      
    case OP_MAIN_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY);
	redraw = REDRAW_FULL;
      }
      break;
      
    case OP_MAIN_GROUP_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY|SENDGROUPREPLY);
	redraw = REDRAW_FULL;
      }
      break;
      
    case OP_MAIN_LIST_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY|SENDLISTREPLY);
	redraw = REDRAW_FULL;
      }
      break;
      
    case OP_MAIN_MAIL:
      ci_send_message(0, 0, 0);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_PIPE_MESSAGE:
      if (Context->vcount) {
	if (ci_pipe_message(Context->hdrs[*Context->v2r[Context->current]]))
	  redraw = REDRAW_FULL;
      }
      break;
      
    case OP_MAIN_PRINT_MESSAGE:
      if (Context->vcount) {
	if (!option(OPTASKPRINT) ||
	    (option(OPTASKPRINT) && mutt_yesorno("Print message?", 0)))
	  mutt_print_message(Context->hdrs[*Context->v2r[Context->current]]);
      }
      break;
      
    case OP_MAIN_FLAG_MESSAGE:
#define TOGGLE_FLAG(x) (x)->flagged = !(x)->flagged; (x)->changed = 1
      
      if (Context->vcount) {
	TOGGLE_FLAG(Context->hdrs[*Context->v2r[Context->current]]);
	if (menu == MENU_MAIN)
	  redraw = REDRAW_CURRENT;
      }
      break;
      
    case OP_MAIN_MARK_AS_NEW:
      if (Context->vcount) {
	if (Context->hdrs[*Context->v2r[Context->current]]->read || Context->hdrs[*Context->v2r[Context->current]]->old) {
	  Context->hdrs[*Context->v2r[Context->current]]->read = 0;
	  Context->hdrs[*Context->v2r[Context->current]]->old = 0;
	  Context->hdrs[*Context->v2r[Context->current]]->changed = 1;
	  Context->new++;
	  if (menu == MENU_MAIN) redraw = REDRAW_CURRENT|REDRAW_STATUS;
	}
      }
      break;
      
#ifdef _PGPPATH
    case OP_MAIN_FORGET_PASSPHRASE:
      pgp_void_passphrase();
      mutt_error("PGP passphrase forgotten.");
      break;
#endif /* _PGPPATH */

    case OP_MAIN_BROWSE_URL:
      if (Context->vcount) {
	ci_browse_url(Context->hdrs[*Context->v2r[Context->current]]);
	redraw = REDRAW_FULL;
      }
      break;

    default:
      if (menu == MENU_MAIN) {
	mutt_error("Key is not bound.  Press '?' for help.");
	beep();
      }
    }
    
    if (menu == MENU_PAGER) {
      menu = MENU_MAIN;
      Context->weed = 1; /* turn header weeding back on. */
      redraw = REDRAW_FULL;
    }
    
    if (done) break;
  }
}
