/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#define MAIN_C 1

#include "mutt.h"
#include "mutt_curses.h"
#include "rfc822.h"
#include "sort.h"
#include "version.h"
#include "copyright.h"

#include <string.h>
#include <stdlib.h>
#include <locale.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <signal.h>

#ifdef USE_SETGID
#include <grp.h>
#endif /* USE_SETGID */

#ifdef NCURSES_MOUSE_VERSION
mmask_t Mousemask;
#endif

void mutt_exit (int code)
{
  endwin();
  exit(code);
}

int mutt_system (const char *s)
{
  struct sigaction act;
  int rc;

#ifdef USE_SETGID
  SETGID(User_gid);
#endif /* USE_SETGID */

  /* restore default signal actions */
  memset(&act, 0, sizeof(struct sigaction));
  act.sa_handler = SIG_DFL;
  sigaction(SIGWINCH, &act, 0);
  sigaction(SIGTERM, &act, 0);
  sigaction(SIGHUP, &act, 0);
  sigaction(SIGINT, &act, 0);
  sigaction(SIGTSTP, &act, 0);
  sigaction(SIGCONT, &act, 0);

  rc = system(s);

  ci_signal_init(); /* restore signal handlers */

#ifdef USE_SETGID
  SETGID(Mail_gid);
#endif /* USE_SETGID */

  return(rc);
}

static void mutt_help (void)
{
  puts("To read messages in a folder:\n\
\tmutt [ -nz ] [ -F <muttrc> ] [ -f <folder> ]\n\n\
\t-F <muttrc>\tspecify an alternate muttrc file\n\
\t-z\t\texit immediately if there are no messages in the mailbox\n\
\t-n\t\tcauses Mutt not to read the system Muttrc\n\n\
To compose a new mail message:\n\
\tmutt [ -n ] [ -i <file> ] [ -s <subj> ] [ -c <addr> ] <addr> [ ... ]\n\n\
\t-c <address>\tspecify a carbon-copy (CC) address\n\
\t-i <file>\tspecify a file which Mutt should include in the reply\n\
\t-s <subj>\tspecify a subject (must be in quotes if it has spaces)\n\n\
To recall a postponed message:\n\
\tmutt [ -n ] [ -F <muttrc> ] -p\n\n\
To display the version number and compile-time definitions:\n\
\tmutt -v");
  exit(0);
}

static void show_version (void)
{
  struct utsname uts;

  puts(Copyright);
  puts(Mutt_version);
  uname(&uts);
#ifndef _AIX
  printf("System info: %s %s", uts.sysname, uts.release);
#else
  printf("System info: %s %s.%s", uts.sysname, uts.version, uts.release);
#endif /* _AIX */

#ifdef NCURSES_VERSION
  printf(" (using ncurses %s)", NCURSES_VERSION);
#endif

  puts("\n\nCompile time definitions:");

#ifdef HOSTNAME
  printf("#define HOSTNAME \"%s\"\n", HOSTNAME);
#endif

#ifdef USE_DOMAIN
  puts("#define USE_DOMAIN");
#endif

#ifdef DOMAIN
  printf("#define DOMAIN \"%s\"\n", DOMAIN);
#endif

#ifdef HIDDEN_HOST
  puts("#define HIDDEN_HOST");
#endif

#ifdef _PGPPATH
  printf("#define _PGPPATH \"%s\"\n", _PGPPATH);
#endif

#ifdef HOMESPOOL
  puts("#define HOMESPOOL");
#endif

#ifdef MAILPATH
  printf("#define MAILPATH \"%s\"\n", MAILPATH);
#endif

#ifdef WORLD_WRITABLE_SPOOL
  puts("#define WORLD_WRITABLE_SPOOL");
#endif

#ifdef USE_SETGID
  puts("#define USE_SETGID");
#endif

#ifdef USE_POP
  puts("#define USE_POP");
#endif

#ifdef HAVE_COLOR
  puts("#define HAVE_COLOR\n");
#endif

  exit(0);
}

static void start_curses (void)
{
  initscr();
  scrollok(stdscr, TRUE);
#ifdef HAVE_COLOR
  ci_start_color();
#endif
  keypad(stdscr, TRUE);
  cbreak();
  noecho();

#ifdef NCURSES_MOUSE_VERSION
  if (option(OPTXTERMMOUSE)) {
    Mousemask = mousemask(BUTTON1_CLICKED|BUTTON1_DOUBLE_CLICKED, 0);
    dprint(1, (debugfile, "main(): mousemask = %d.\n", (int)Mousemask));
  }
#endif /* NCURSES_MOUSE_VERSION */
  
  ci_signal_init();
  SETCOLOR(MT_COLOR_NORMAL);
}

static void ci_endwin (const char *msg)
{
  move(LINES-1, COLS-1);
  refresh();
  fputc('\n', stdout);
  endwin();
  if (msg) puts(msg);
}

int main (int argc, char **argv)
{
  char folder[_POSIX_PATH_MAX] = { 0 };
  char *subject=0;
  char *includeFile=0, *ccList=0;
  int i;
  int nosysrc=0;
  int sendflags=0;
  int ignore_empty=0;
  extern char *optarg;
  extern int optind;

  SRAND(time(0));
  setlocale(LC_CTYPE, "");
  umask(077);

  strcpy(Muttrc, "~/.muttrc");

  /* Check for a batch send. */
  if (!isatty(0)) sendflags = SENDBATCH;

  /* Process the arguments */
  while ((i = getopt(argc, argv, "F:f:c:d:s:i:hnpvz")) != EOF)
    switch(i) {
    case 'F':
      strfcpy(Muttrc, optarg, sizeof(Muttrc));
      break;
    case 'f':
      strfcpy(folder, optarg, sizeof(folder));
      break;
    case 'c':
      ccList = optarg;
      break;
    case 'd':
#ifdef DEBUG
      debuglevel = atoi(optarg);
      printf("Debugging at level %d.\n", debuglevel);
#else
      printf("DEBUG was not defined during compilation.  Ignored.\n");
#endif
      break;
    case 'i':
      includeFile = optarg;
      break;
    case 'n':
      nosysrc = 1;
      break;
    case 'p':
      sendflags |= SENDPOSTPONED;
      break;
    case 's':
      subject = optarg;
      break;
    case 'v':
      show_version();
    case 'z':
      ignore_empty=1;
      break;
    default:
      mutt_help();
    }

  ci_init_color();
  keymap_init(); /* must come before mutt_init */
  if (mutt_init(nosysrc) == -1) ci_any_key_to_continue((char *)0);

  if (sendflags & SENDPOSTPONED) {
    start_curses();
    ci_send_message((METAENV *)0, (char *)0, SENDPOSTPONED);
    ci_endwin((char *)0);
  }
  else if (sendflags || optind < argc) {
    METAENV *env = mutt_new_metaenv();

    env->env = mutt_new_envelope();
    env->env->subject = safe_strdup(subject);
    for (i=optind; i<argc; i++)
      rfc822_parse_adrlist(&env->env->to, argv[i], "@");
    if (!env->env->to) {
      puts("No recipients specified.");
      exit(0);
    }
    if (ccList) rfc822_parse_adrlist(&env->env->cc, ccList, "@");
    if (!(sendflags & SENDBATCH)) start_curses();
    ci_send_message(env, includeFile, sendflags);
    if (!(sendflags & SENDBATCH)) ci_endwin((char *)0);
  }
  else {
    struct stat st;
#ifdef USE_SETGID
    struct group *grp = getgrnam("mail");

    if (!grp) {
      puts("Error!  Couldn't find group entry for \"mail\".");
      exit(0);
    }
    User_gid = getgid();
    Mail_gid = grp->gr_gid;
    if (SETGID(Mail_gid) != 0) {
      dprint(1,(debugfile, "main(): setegid() returned -1; errno = %d.\n", errno));
      puts("Error!  Could not change gid to mail!");
      exit(0);
    }
#endif /* USE_SETGID */

    if (!folder[0]) {
      strfcpy(folder, Spoolfile, sizeof(folder));
      /* if the spool folder doesn't exist, try to create it. */
      if (stat(folder, &st) != 0 && errno == ENOENT) {
	FILE *f = fopen(folder, "w");
	
	if (f) fclose(f);
      }
    }
    if (ignore_empty) {
      /* check to see if there are any messages in the folder */
      mutt_expand_path(folder);
      if (stat(folder, &st) != 0) {
	puts(errno == ENOENT ? "Mailbox does not exist." : "Could not stat that mailbox!");
	exit(0);
      }
      if (st.st_size == 0) {
	puts("Mailbox is empty.");
	exit(0);
      }
    }
    start_curses();
    if ((Context = mutt_open_mailbox(folder, (CONTEXT *)0, 0)) != 0) {
      ci_main();
      ci_endwin((char *)0);
    }
    else
      ci_endwin("Unable to open mailbox.");
  }
  exit(0);
}
