/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 


#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define ISHEADER(x) ((x == MT_COLOR_HEADER) || (x == MT_COLOR_SUBJECT) || (x == MT_COLOR_FROM))

int ci_simple_pager (const char *fname, int do_color)
{
  unsigned char buffer[LONG_STRING];
  unsigned char *pending=0;
  int i, ch = 0, rc = -1, chars = 0, y, x;
  int last_match = -1; /* last line that matched the search pattern */
  long **offset=0, offset_max=0, offset_len=0;
  int lines=0, curline=0;
  FILE *fp=NULL;
  struct stat sb;

  /* These are used for color support. */
  int attr = 0;
  int **class = 0; /* line classification */
  int err;
#ifdef HAVE_REGCOMP
  regex_t quote_re; /* how to find quoted lines in the body of a message */
  regex_t re;
#elif HAVE_REGCMP
  char *quote_re;
  char *re;
#endif

  static char searchbuf[STRING];
  int searching=0, savecurline=0;

  if ((fp = fopen(fname, "r")) == NULL) {
    mutt_error("Could not open file %s!", fname);
    return(-1);
  }
  stat(fname, &sb);
  unlink(fname);

  /* compile the expression for finding quoted lines */
  if ((err = REGCOMP(quote_re, QuoteString)) != 0) {
#ifdef HAVE_REGCOMP
    regerror(err, &quote_re, buffer, sizeof(buffer));
    mutt_error(buffer);
#else
    mutt_error("quote_regexp contains an invalid expression.");
#endif
    REGFREE(quote_re);
    fclose(fp);
    return(-1);
  }

  clear();

  while (ch != -1) {
    while (lines < LINES-1) {
      if (!pending) {

	if (offset_max == offset_len) {
	  if (offset == 0) {
	    offset = safe_malloc(sizeof(long *)*((offset_max = LINES)));
	    class = safe_malloc(sizeof(int *) * offset_max);
	  }
	  else {
	    offset_max += LINES;
	    safe_realloc((void **)&offset, sizeof(long *) * offset_max);
	    safe_realloc((void **)&class, sizeof(int *) * offset_max);
	  }
	}

	if (curline == offset_len) {
	  offset[curline] = safe_malloc(sizeof(long));
	  *offset[curline] = ftell(fp);

	  class[curline] = safe_malloc(sizeof(int));
	  *class[curline] = -1;
	  offset_len++;
	}

	if (fgets(buffer, sizeof(buffer), fp) == NULL) break;	      

	pending = buffer;
	
	 /* Attempt to figure out what "class" this line is. */
	if (*class[curline] == -1) {
	  if (curline == 3 ||
	      (curline > 3 && ISHEADER(*class[curline-1]) && buffer[0] != '\n')) {
	    if (strncasecmp("subject:", buffer, 8) == 0)
	      *class[curline] = MT_COLOR_SUBJECT;
	    else if (strncasecmp("from:", buffer, 5) == 0)
	      *class[curline] = MT_COLOR_FROM;
	    else
	      *class[curline] = MT_COLOR_HEADER;
	  }
	  else if (strncmp(buffer, "[Attachment #", 13) == 0)
	    *class[curline] = MT_COLOR_ATTACHMENT;
	  else if (strcmp("-- \n", buffer) == 0)
	    *class[curline] = MT_COLOR_SIGNATURE;
	  else if (curline > 0 && *class[curline-1] == MT_COLOR_SIGNATURE)
	    *class[curline] = MT_COLOR_SIGNATURE;
	  else if (REGEXEC(quote_re, buffer) == 0)
	    *class[curline] = MT_COLOR_QUOTED;
	  else
	    *class[curline] = MT_COLOR_NORMAL;
	}

	attr = *class[curline];
      }

      if (searching) {
	pending = 0;
	if (REGEXEC(re, buffer) == 0) {
	  clear();
	  searching = 0;
	  fseek(fp, *offset[curline], 0); /* reposition at the begin of line */
	  last_match = curline;
#ifdef HAVE_REGCOMP
	  regfree(&re);
#endif
	  continue;
	}
	curline++;
	continue;
      }

      if (do_color) SETCOLOR(attr);
      chars = 0;
      while (*pending && lines<LINES-1) {
	if (chars == COLS) {
	  lines++;
	  chars = 0;
	  /* If the last character on the line is right at the edge of
	   * the screen, don't print the newline, or else we'll end up
	   *  with a blank line.
	   */
	  if (*pending == '\n') pending++;
	  continue;
	}
	if (*pending == '\n') {
	  addch(*pending);
	  chars = 0;
	  lines++;
	}
#define MUTT_TABSTOP 8
#define next_tabstop (chars > (COLS/8)*8 ? 0 : (((chars/MUTT_TABSTOP)+1)*MUTT_TABSTOP))
	else if (*pending == '\t') {
	  i = next_tabstop;
	  if (i == 0) {
	    addch('\n');
	    lines++;
	    chars = 0;
	  }
	  else
	    for (i=i-chars; i > 0; i--, chars++) addch(' ');
	}
	else if (IsPrint(*pending)) {
	  addch(*pending);
	  chars++;
	}
	else if (iscntrl(*pending)) {
	  if (chars > COLS-1) {
	    addch('\n');
	    chars = 0;
	    lines++;
	    continue;
	  }
	  addch('^');
	  addch(*pending+'@');
	  chars += 2;
	  if (*pending == ctrl('L')) {
	    pending++;
	    break;
	  }
	}
	else
	  addch('?'); /* unprintable character! */
	pending++;
      }

      if (*pending == 0) {
	/* finished with this line */
	curline++;
	pending = 0;
      }
    }

    if (do_color) SETCOLOR(MT_COLOR_NORMAL);

    if (searching) {
      CLEARLINE(LINES-1);
      attrset(A_REVERSE);
      addstr("Not found!");
      attrset(A_NORMAL);
      searching=0;
      curline=savecurline;
      fseek(fp, *offset[curline-1], 0);
      REGFREE(re);
    }
    else {
      if (!option(OPTFLOATINGPROMPT))
	move(LINES-1, 0);
      else {
	getyx(stdscr, y, x);
	CLEARLINE(y);
      }

      attron(A_REVERSE);
      if (ftell(fp) < sb.st_size-1)
	printw("More (you've seen %d%%)", 100*ftell(fp)/sb.st_size);
      else
	addstr("(END)");
      attroff(A_REVERSE);
    }

    refresh();
    ch = ci_getch();

#ifdef NCURSES_MOUSE_VERSION
    if (ch == KEY_MOUSE) {
      MEVENT e;
      
      getmouse(&e);
      if (e.bstate & BUTTON1_CLICKED) {
	if (e.y <= LINES/2)
	  ch = OP_PAGER_NEXT_PAGE;
	else
	  ch = OP_PAGER_PREVIOUS_PAGE;
      }
      else
	continue;
      }
    else
#endif
      ch = keymap_lookup(ch, MENU_PAGER);

    rc = ch;

    switch (ch) {
    case OP_PAGER_EXIT:
      rc = -1;
      ch = -1;
      break;

    case OP_PAGER_NEXT_PAGE:
      if (ftell(fp) < sb.st_size-1) {
	lines = 0;
	clear();
      }
      else if (option(OPTPAGERSTOP)) {
	/* emulate "less -q" and don't go on to the next message. */
	beep();
      }
      else {
	/* end of the current message, so display the next message. */
	rc = OP_MAIN_NEXT_UNDELETED;
	ch = -1;
      }
      break;

    case OP_PAGER_PREVIOUS_PAGE:
      if (curline != 0) {
	clear();
	curline -= 2*LINES-2;
	if (curline < 0) curline = 0;
	fseek(fp, *offset[curline], 0);
	lines = 0;
      }
      break;

    case OP_PAGER_NEXT_LINE:
      getyx(stdscr, y, x);
      CLEARLINE(y);
      if (ftell(fp) < sb.st_size-1)
	lines--;
      else
	beep();
      break;

    case OP_PAGER_PREVIOUS_LINE:
      clear();
      if ((curline -= LINES) < 0) curline = 0;
      fseek(fp, *offset[curline], 0);
      lines = 0;
      break;

    case OP_PAGER_TOP:
      if (curline > LINES) {
	clear();
	lines=0;
	fseek(fp, 0, 0);
	curline=0;
      }
      else
	beep();
      break;

    case OP_PAGER_SEARCH:
      if (ci_get_field("Search: ", searchbuf, sizeof(searchbuf), 0) != 0 || !searchbuf[0]) break;

      if ((err = REGCOMP(re, searchbuf)) != 0) {
#ifdef HAVE_REGCOMP
	char error[STRING];
	
	regerror(err, &re, error, sizeof(error));
	mutt_error(error);
#else
	mutt_error("Error in regular expression.");
#endif
	REGFREE(re);
	searching=0;
	fseek(fp, *offset[curline-1], 0);
      }
      else {
	searching=1;
	lines=0;
	savecurline = curline;
	if (last_match == -1) {
	  if ((curline -= LINES-2) < 0) curline = 0;
	}
	else
	  curline = last_match + 1;
	fseek(fp, *offset[curline], 0);
      }
      break;

    case OP_PAGER_HELP:
      ci_help(MENU_PAGER);

      lines=0;
      curline-=LINES-2;
      if (curline < 0) curline=0;
      fseek(fp, *offset[curline], 0);
      break;

    default:
      ch = -1;
    }
  }
  fclose(fp);
  for (i=0; i<offset_len; i++) {
    safe_free((void **)&offset[i]);
    safe_free((void **)&class[i]);
  }
  safe_free((void **)&offset);
  safe_free((void **)&class);
  REGFREE(quote_re);
  return (rc != -1 ? rc : 0);
}
