/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"

#include <unistd.h>
#include <string.h>

extern ENVELOPE *parse_header();

static struct {
  CONTEXT *context;
} LOCAL = { 0 };

void mutt_postpone_message (const char *tempfile,
			    METAENV *env,
			    const char *msgid,
			    BODY *attach)
{
  time_t now;
  FOLDER *f;
  int len;
  char buf[LONG_STRING];
  FILE *t;

  if ((f = folder_open(Postponed, "a")) == 0) return;
  if (lock_file(f, TRUE) != 0) {
    folder_close(&f);
    mutt_error("Could not lock %s!", Postponed);
    return;
  }
  fseek(f->fp, 0, 2);
  
  now = time(0);
  fprintf(f->fp, "From %s %s", Username, ctime(&now));
  mutt_write_rfc822_header(f->fp, env, NULL, -1);

  /*
   * if this was a reply of some sort, <msgid> contians the
   * Message-ID: of message replied to.  Save it using a special X-Mutt-
   * header so it can be picked up if the message is recalled at a later
   * point in time.  This will allow the message to be marked as replied if
   * the same mailbox is still open.
   */
  if (msgid) fprintf(f->fp, "X-Mutt-References: %s\n", msgid);

  /*
   * Save the attachments to this message using
   * X-Mutt-Attachment: <filename> [ <description> ]
   */
  while (attach) {
    fprintf(f->fp, "X-Mutt-Attachment: %s", attach->filename);
    if (attach->description) {
      fputc(' ', f->fp);
      fputs(attach->description, f->fp);
    }
    fputc('\n', f->fp);
    attach = attach->next;
  }

  fputc('\n', f->fp); /* finish off the header. */
  if ((t = fopen(tempfile, "r")) != 0) {
    while ((len = fread(buf, 1, sizeof(buf), t)) > 0)
      fwrite(buf, 1, len, f->fp);
    fclose(t);  
    fputc('\n', f->fp);
  }
  unlink(tempfile);
  unlock_file(f);
  folder_close(&f);
}

static void post_make_entry (char *s, int entry, int slen)
{
  char buf[SHORT_STRING] = { 0 };

  mutt_write_address(buf, sizeof(buf), LOCAL.context->hdrs[entry]->to);
  snprintf(s, slen, "%2d  %-20.20s     %s", entry+1, buf,
	   LOCAL.context->hdrs[entry]->subject);
}

static HEADER *select_msg (void)
{
  MUTTMENU *menu;
  int i, done=0, r=-1;

  menu = mutt_newMenu("Postponed Messages");
  menu->makeEntry = post_make_entry;
  menu->menu = MENU_POST;
  menu->max = LOCAL.context->msgcount;

  while (!done) {
    i = mutt_menuLoop(menu);
    switch(i) {
    case OP_GENERIC_DELETE_ENTRY:
      LOCAL.context->hdrs[menu->current]->deleted = 1;
      LOCAL.context->hdrs[menu->current]->changed = 1;
      break;
    case OP_GENERIC_UNDELETE_ENTRY:
      LOCAL.context->hdrs[menu->current]->deleted = 0;
      LOCAL.context->hdrs[menu->current]->changed = 0;
      break;
    case OP_GENERIC_SELECT_ENTRY:
      r = menu->current;
      done=1;
      break;
    case OP_GENERIC_EXIT:
      done=1;
      break;
    }
  }

  mutt_menuDestroy(&menu);
  return(r > -1 ? LOCAL.context->hdrs[r] : NULL);
}

int ci_get_postponed_message (FILE *temp,
			      METAENV *env,
			      HEADER **cur,
			      BODY **attach)
{
  HEADER *h;
  long bytes;
  BODY *ptr = 0;

  *cur = 0;
  *attach = 0;

  if ((LOCAL.context = mutt_open_mailbox(Postponed, NULL, M_NOSORT)) == 0)
    return(-1);
  
  if (LOCAL.context->msgcount == 1) {
    /* only one message, so just use that one. */
    h = LOCAL.context->hdrs[0];
  }
  else if ((h = select_msg()) == (HEADER *)0) {
    mutt_close_mailbox(LOCAL.context);
    safe_free((void **)&LOCAL.context);
    return(-1);
  }

  fseek(LOCAL.context->folder->fp, h->offset, 0);
  env->env = parse_header(LOCAL.context->folder->fp,
			  NULL,
			  env->userhdrs,
			  MAXUSERHDRS,
			  "@");
  bytes = h->length - (ftell(LOCAL.context->folder->fp) - h->offset);
  mutt_copy_bytes(LOCAL.context->folder->fp, temp, bytes);

  /* finished with this message, so delete it. */
  h->deleted = 1;
  h->changed = 1;

  mutt_close_mailbox(LOCAL.context);
  safe_free((void **)&LOCAL.context);

  if (Context) {
    int i, j;

    for (i=0; env->userhdrs[i]; i++) {
      if (strncmp("X-Mutt-References:", env->userhdrs[i], 18) == 0) {
	char *p = env->userhdrs[i] + 18;

	while (*p && (*p == ' ' || *p == '\t')) p++;

	for (j=0; j < Context->msgcount; j++) {
	  if (Context->hdrs[j]->message_id &&
	      strcmp(Context->hdrs[j]->message_id, p) == 0) {
	    *cur = Context->hdrs[j];
	    break;
	  }
	}

	/* Remove the X-Mutt-References: header field. */
	safe_free((void **)&env->userhdrs[i]);
	for (j=i+1; env->userhdrs[j]; j++)
	  env->userhdrs[j-1] = env->userhdrs[j];

	return(SENDREPLY);
      }
      else if (strncmp("X-Mutt-Attachment:", env->userhdrs[i], 18) == 0) {
	char *p = skip_whitespace(env->userhdrs[i]+18);
	char *q;

	if ((q = strchr(p, ' ')) != 0) *q++ = 0;

	if (*attach) {
	  ptr->next = mutt_gen_attach(p);
	  if (ptr->next) ptr = ptr->next;
	}
	else
	  *attach = ptr = mutt_gen_attach(p);
	if (q) ptr->description = safe_strdup(q);

	safe_free((void **)&env->userhdrs[i]);
	for (j=i+1; env->userhdrs[j]; j++)
	  env->userhdrs[j-1] = env->userhdrs[j];
      }
    }
  }
  return(0);
}
